import pickle
import typing

import pytest
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA_sklearn
import numpy as np
import scipy.stats

from scalib import ScalibError
from scalib.modeling import LDAClassifier, MultiLDA, Lda, LdaAcc

from utils_test import get_rng


def is_parallel(x, y):
    z = np.abs(x.dot(y))
    z2 = np.linalg.norm(x) * np.linalg.norm(y)
    return np.allclose(z, z2, rtol=1e-3)


def parallel_factor(x, y):
    """Return 1 if x and y are parallel and in the same direction, and -1 is
    they are in opposite directions.
    """
    x = x / np.linalg.norm(x)
    y = y / np.linalg.norm(y)
    return np.dot(x, y)


def test_lda_pickle():
    # np.set_printoptions(threshold=np.inf)
    ns = 10
    n_components = 2
    nc = 4
    n = 5000

    rng = get_rng()
    m = rng.integers(0, 100, (nc, ns))
    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    dumped_lda = pickle.dumps(lda)
    lda = pickle.loads(dumped_lda)

    lda_ref = LDA_sklearn(solver="eigen", n_components=n_components)
    lda_ref.fit(traces, labels)

    # Project traces with SCALib
    from scalib.config import get_config

    ptraces = lda.mlda.project(traces, get_config())[0]
    # Project traces woth sklearn
    ptraces_sklearn = lda_ref.transform(traces)
    projections_similar = all(
        [is_parallel(a, b) for a, b in zip(ptraces.T, ptraces_sklearn.T)]
    )
    assert projections_similar, (ptraces, ptraces_sklearn)

    # generate means and cov in subspace
    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    means_check = np.zeros((nc, n_components))
    for i in range(nc):
        I = np.where(labels == i)[0]
        means_check[i, :] = np.mean(traces_t[I, :], axis=0)
    traces_t = traces_t - means_check[labels, :]
    cov_check = np.cov(traces_t.T)

    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    prs = lda.predict_proba(traces)
    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    prs_ref = np.zeros((len(traces), nc))
    for x in range(nc):
        prs_ref[:, x] = scipy.stats.multivariate_normal.pdf(
            traces_t, means_check[x], cov_check
        )
    prs_ref = (prs_ref.T / np.sum(prs_ref, axis=1)).T

    assert np.allclose(prs, prs_ref, rtol=1e-2)


def test_lda():
    # np.set_printoptions(threshold=np.inf)
    ns = 10
    n_components = 2
    nc = 4
    n = 500

    rng = get_rng()
    m = rng.integers(0, 100, (nc, ns))
    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    pr = lda.predict_proba(traces)
    print(pr.shape)

    lda_ref = LDA_sklearn(solver="eigen", n_components=n_components)
    lda_ref.fit(traces, labels)

    # Project traces with SCALib
    from scalib.config import get_config

    ptraces = lda.mlda.project(traces, get_config())[0]
    # Project traces woth sklearn
    ptraces_sklearn = lda_ref.transform(traces)
    projections_similar = all(
        [is_parallel(a, b) for a, b in zip(ptraces.T, ptraces_sklearn.T)]
    )
    assert projections_similar, (ptraces, ptraces_sklearn)

    # Verify value of means by accessor
    ref_means = lda_ref.means_
    lda_means = lda.get_mus()
    assert np.allclose(ref_means, lda_means, rtol=1e-10)

    # Verify scatter matrix by accessor
    lda_scat = lda.get_sw() / n
    cov_ref = lda_ref.covariance_
    assert np.allclose(lda_scat, cov_ref, rtol=1e-5)

    # Not point of comparison with sklearn, but we here call
    # the accessor for the inter-class scatter matrix just to verify
    # that its worling.
    smat = lda.get_sb()

    # We can't do much more since sklearn has no way to reduce dimensionality of LDA.
    # e.g., comparing probas will fail


def test_lda_noproj():
    ns = 10
    n_components = 3
    nc = 4
    n = 500

    rng = get_rng()
    m = rng.integers(0, 100, (nc, ns))
    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components)
    lda.fit_u(traces, labels, 1)
    lda.solve()

    lda_ref = LDA_sklearn(solver="eigen", n_components=n_components)
    lda_ref.fit(traces, labels)

    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    prs = lda.predict_proba(traces)
    prs_ref = lda_ref.predict_proba(traces)

    assert np.allclose(prs, prs_ref, rtol=1e-2, atol=1e-3)


def test_lda_fail_bad_traces():
    # Issue #56
    n = 100
    ns = 6
    nc = 4
    lda = LDAClassifier(nc, 3)
    rng = get_rng()
    traces_bad = rng.integers(0, 1, (n, ns), dtype=np.int16)
    y = rng.integers(0, nc, n, dtype=np.uint16)
    lda.fit_u(traces_bad, y, 0)
    with pytest.raises(ScalibError):
        lda.solve()


def test_multilda():
    rng = get_rng()
    x = rng.integers(0, 256, (5000, 50), dtype=np.int16)
    y = rng.integers(0, 256, (5000, 5), dtype=np.uint16)
    pois = [list(range(7 * i, 7 * i + 10)) for i in range(5)]
    lda = MultiLDA(5 * [256], 5 * [3], pois)
    lda.fit_u(x, y)
    lda.solve()
    x = rng.integers(0, 256, (20, 50), dtype=np.int16)
    _ = lda.predict_proba(x)


def multi_lda_data_indep(ns, nc, nv, n, n_batches, rng):
    y = [rng.integers(0, nc, (n, nv), dtype=np.uint16) for _ in n_batches]
    traces = [
        rng.integers(-(2**15), 2**15, (n, ns), dtype=np.uint16) for _ in n_batches
    ]
    return traces, y


def multi_lda_gen_pois_overlap(rng, ns, nv, npois):
    pois = np.tile(np.arange(ns), (nv, 1))
    rng.shuffle(pois, axis=1)
    return pois[:, :npois]


def multi_lda_gen_pois_consec(nv, npois, gap=0):
    return np.array(
        [np.arange(i * (npois + gap), i * (npois + gap) + npois) for i in range(nv)]
    )


def multi_lda_gen_indep_overlap(rng, ns, nc, nv, npois, n, n_batches, maxl=2**15, **_):
    pois = np.tile(np.arange(ns), (nv, 1))
    rng.shuffle(pois, axis=1)
    pois = pois[:, :npois]
    y = [rng.integers(0, nc, (n, nv), dtype=np.uint16) for _ in range(n_batches)]
    traces = [
        rng.integers(-maxl, maxl, (n, ns), dtype=np.int16) for _ in range(n_batches)
    ]
    return pois, traces, y


def ldaacc_sklearn_compare(nc, nv, p, pois, traces, x, **_):
    traces_all = np.vstack(traces)
    x_all = np.vstack(x)
    ldas_ref = [LDA_sklearn(solver="eigen", n_components=p) for _ in range(nv)]
    ldaacc = LdaAcc(nc=nc, pois=pois)
    # Fit SKlean
    for i, (pi, xi) in enumerate(zip(pois, x_all.T)):
        ldas_ref[i].fit(traces_all[:, pi].reshape([traces_all.shape[0], len(pi)]), xi)
    # Fit the scalib lda
    for t, y in zip(traces, x):
        ldaacc.fit_u(t, y)
    # Check the mus matrix
    for mu, mu_ref in zip(ldaacc.get_mus(), [lda.means_ for lda in ldas_ref]):
        assert np.allclose(mu, mu_ref)
    ## Check the within scatter matrix
    for sw, sw_ref in zip(ldaacc.get_sw(), [lda.covariance_ for lda in ldas_ref]):
        assert np.allclose(sw / traces_all.shape[0], sw_ref, rtol=1e-5)
    ## No point of comparison for the between scatter matrix, but call the accessor to check that the function works
    _ = ldaacc.get_sb()
    # We can't do much more since sklearn has no way to reduce dimensionality of LDA.
    # e.g., comparing probas will fail


def test_ldaacc_sklearn_compare():
    base_case = dict(ns=2, nc=2, nv=1, npois=1, n=10, n_batches=1, p=1, test_lp=True)
    cases = [
        base_case,
        base_case | dict(n=5, n_batches=2),
        base_case | dict(ns=3, n=20, maxl=100),
        base_case | dict(ns=20, nc=4, nv=4, n=20, npois=5, n_batches=3, p=2),
        base_case | dict(ns=100, nc=256, nv=2, npois=20, n=5000, n_batches=5, p=4),
        base_case | dict(ns=1000, nc=4, nv=10, n=500, n_batches=5),
    ]
    for case in cases:
        rng = get_rng(**case)
        print(80 * "#" + "\ncase:", case)
        pois, traces, x = multi_lda_gen_indep_overlap(rng, **case)
        ldaacc_sklearn_compare(pois=pois, traces=traces, x=x, **case)


def multi_lda_compare(nc, nv, p, pois, traces, x, test_lp=False, **_):
    ncs = [nc for _ in range(nv)]
    ps = [p for _ in range(nv)]
    multi_lda = MultiLDA(ncs, ps, pois=pois)
    multi_lda3 = LdaAcc(pois=pois, nc=nc)
    for t, y in zip(traces, x):
        multi_lda.fit_u(t, y)
        multi_lda3.fit_u(t, y)
    for mus, mus3 in zip(multi_lda.get_mus(), multi_lda3.get_mus()):
        assert np.allclose(mus, mus3)
    for sb, sb3 in zip(multi_lda.get_sb(), multi_lda3.get_sb()):
        assert np.allclose(sb, sb3)
    for sw, sw3 in zip(multi_lda.get_sw(), multi_lda3.get_sw()):
        assert np.allclose(sw, sw3)
    multi_lda.solve(done=False)
    multi_lda3 = Lda(multi_lda3, p=p)
    for t, y in zip(traces, x):
        probas = np.array(multi_lda.predict_proba(t))
        probas3 = multi_lda3.predict_proba(t)
        assert np.allclose(probas, probas3)
        if test_lp:
            lp = multi_lda3.predict_log2_proba_class(t, y)
            lp3 = np.log2(
                probas3[
                    np.arange(probas3.shape[0])[:, np.newaxis],
                    np.arange(probas3.shape[1])[np.newaxis, :],
                    y.T,
                ]
            )
            assert np.allclose(lp, lp3)


def test_multi_lda_compare():
    base_case = dict(ns=2, nc=2, nv=1, npois=1, n=10, n_batches=1, p=1, test_lp=True)
    cases = [
        base_case,
        base_case | dict(n=5, n_batches=2),
        base_case | dict(ns=3, n=20, maxl=100),
        base_case | dict(ns=20, nc=4, nv=4, npois=5, n_batches=3, p=2),
        base_case | dict(ns=100, nc=256, nv=2, npois=20, n=500, n_batches=5, p=4),
        base_case | dict(ns=1000, nc=4, nv=10, n_batches=5),
    ]
    for case in cases:
        rng = get_rng(**case)
        print(80 * "#" + "\ncase:", case)
        pois, traces, x = multi_lda_gen_indep_overlap(rng, **case)
        multi_lda_compare(pois=pois, traces=traces, x=x, **case)


def test_simple_multi_lda_compare():
    nc = 2
    nv = 1
    pois = [np.array([0, 1])]
    p = 1
    traces = [np.array([[1, 0], [0, 1], [0, 0], [4, 4]], dtype=np.int16)]
    x = [np.array([[0], [0], [1], [1]]).astype(np.uint16)]
    lda = LDAClassifier(nc, p)
    lda.fit_u(traces[0], x[0][:, 0])
    lda.solve()
    multi_lda_compare(nc=nc, nv=nv, p=p, pois=pois, traces=traces, x=x)


def test_seq_multi_lda_compare():
    cases = [
        dict(nv=100, npois=5, nc=4, n=1000, p=2),
        dict(nv=2000, npois=5, nc=2, n=30, p=1),
    ]
    for case in cases:
        print(80 * "#" + "\ncase:", case)
        nv = case["nv"]
        npois = case["npois"]
        nc = case["nc"]
        n = case["n"]
        p = case["p"]
        rng = get_rng(**case)
        _, traces, x = multi_lda_gen_indep_overlap(
            rng, ns=nv * npois, nc=nc, nv=nv, npois=0, n=n, n_batches=1
        )
        pois = [list(range(i * npois, (i + 1) * npois)) for i in range(nv)]
        multi_lda_compare(nc=nc, nv=nv, p=p, pois=pois, traces=traces, x=x)


def test_multi_lda_pickle():
    ns = 10
    nc = 4
    n = 5000
    rng = get_rng()
    traces = rng.integers(0, 10, (n, ns), dtype=np.int16)
    labels = rng.integers(0, nc, (n, 1), dtype=np.uint16)
    lda_acc = LdaAcc(pois=[list(range(ns))], nc=nc)
    lda_acc.fit_u(traces, labels)
    dumped_lda_acc = pickle.dumps(lda_acc)
    lda_acc2 = pickle.loads(dumped_lda_acc)
    lda = Lda(lda_acc, p=2)
    lda2 = Lda(lda_acc2, p=2)

    dumped_lda = pickle.dumps(lda2)
    lda2 = pickle.loads(dumped_lda)

    prs = lda.predict_proba(traces)
    prs2 = lda2.predict_proba(traces)

    assert np.allclose(prs, prs2)


def multi_lda_select_simple(rng, nv, ns, npois, nv_sel, n_sel, permute=True):
    nc = 4
    n = 100
    pois, (traces,), (labels,) = multi_lda_gen_indep_overlap(
        rng, ns, nc, nv, npois, n, n_batches=1
    )
    lda_acc = LdaAcc(pois=pois, nc=nc)
    lda_acc.fit_u(traces, labels)
    lda_all = Lda(lda_acc, p=1)
    prs_all = lda_all.predict_proba(traces)
    # Validate for a random selection
    for _ in range(n_sel):
        if permute:
            selection = list(rng.permutation(range(nv))[:nv_sel])
        else:
            selection = list(rng.integers(0, nv, (nv_sel,)))
        lda_s_only = lda_all.select_vars(selection)
        prt = lda_s_only.predict_proba(traces)
        assert np.allclose(prs_all[selection, ...], prt)


def test_multi_lda_select():
    cases: list[dict[str, typing.Any]] = [
        dict(nv=5, ns=25, npois=1, nv_sel=5, n_sel=2),
        dict(nv=5, ns=25, npois=1, nv_sel=1, n_sel=2),
        dict(nv=5, ns=25, npois=1, nv_sel=2, n_sel=2),
        dict(nv=5, ns=25, npois=5, nv_sel=2, n_sel=2),
        dict(nv=5, ns=25, npois=15, nv_sel=2, n_sel=1),
        dict(nv=5, ns=25, npois=10, nv_sel=8, n_sel=1, permute=False),
    ]
    for case in cases:
        rng = get_rng(**case)
        multi_lda_select_simple(rng, **case)



from typing import TypeAlias

import numpy as np
import numpy.typing as npt

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils


class Cpa:
    r"""Performs a Correlation Power Attacks (CPA) by computing the pearson
    correlation between the traces associated to a specific intermediate state
    and an arbitrary leakage model [BrierCO04]_ . The correlation metric is computed
    over a range of key guesses, such that the key value maximising the
    correlation absolute value is considered as the correct key guess.

    The intermediate state :math:`y \in [0; nc[` is modeled as a function of
    the value :math:`x \in [0; nc[` and the key guess :math:`k_g \in [0; nc[`
    such that :math:`y=\text{intermediate}(x, k_g)`. Currently, two
    intermediate functions are supported: the bitwise xor and the addition
    modulo `nc`.

    The correlation metric between the leakages samples :math:`L_x` and their
    models :math:`M_y` is computed according to following equation:

    .. math::
        \mathrm{\hat{\rho}(L_x;M_y)} = \dfrac{\hat{\text{cov}}\left( L_x;M_y\right)}{\hat{\sigma}_{L_x}\sigma_{M_y}}

    where

    :math:`\hat{\text{cov}}\left( L_x;M_y\right)` :
        is the unbiased estimation of the covariance between the leakage and
        the models,
    :math:`\hat{\sigma}_{L_x}` :
        is the unbiased estimation of the leakages samples standard deviation.
    :math:`\sigma_{M_y}` :
        is the exact value of the model standard deviation, computed over the
        exhaustive model distribution provided.


    Parameters
    ----------
    nc : int
        Number of possible values for the random variable :math:`X` (e.g., 256 for 8-bit
        target). ``nc`` must be between :math:`2` and :math:`2^{16}` (included).
    kind:
        Addition function between key and label (``Cpa.Xor``, ``Cpa.Add``).
    use_64bit : bool (default False)
        Use 64 bits for intermediate sums instead of 32 bits.
        When using 64-bit sums, SNR can accumulate up to :math:`2^{32}` traces, while when
        32-bit sums are used, the bound is :math:`n_i < 2^{32}/b`, where b is the
        maximum absolute value of a sample rounded to the next power of 2, and
        :math:`n_i` is the maximum number of times a variable can take a given value.
        Concretely, the total number of traces `n` should be at most
        :math:`(nc \cdot 2^{32}/b) - k` , where :math:`k = O(\sqrt{n})`, typ.
        :math:`k>=3*\sqrt{n}`  (see https://mathoverflow.net/a/273060).

    Examples
    --------
    >>> from scalib.attacks import Cpa
    >>> import numpy as np
    >>> # 500 traces of 200 points, 8-bit samples
    >>> traces = np.random.randint(0,256,(500,200),dtype=np.int16)
    >>> # 10 variables on 8 bit (256 classes = 2^8)
    >>> x = np.random.randint(0,256,(500,10),dtype=np.uint16)
    >>> cpa = Cpa(nc=256, kind=Cpa.Xor)
    >>> cpa.fit_u(traces,x)
    >>> hamming_weights = np.array([x.bit_count() for x in range(256)], dtype=np.float64)
    >>> models = np.tile(hamming_weights[np.newaxis,:,np.newaxis], (10, 1, 200))
    >>> cpa_val = cpa.get_correlation(models)

    Notes
    -----
    .. [BrierCO04] "Correlation Power Analysis with a Leakage Model", Eric Brier, Christophe Clavier, Francis Olivier, CHES 2004: 16-29

    """

    IntermediateKind: TypeAlias = _scalib_ext.CpaIntermediateKind
    Xor = IntermediateKind.Xor
    Add = IntermediateKind.Add

    def __init__(self, nc: int, kind: IntermediateKind, use_64bit: bool = False):
        if nc not in range(2, 2**16 + 1):
            raise ValueError(
                "CPA can be computed on max 16 bit variable (and at least 2 classes),"
                f" {nc=} given."
            )
        self._nc = nc
        assert isinstance(kind, self.IntermediateKind)
        self._kind = kind
        self._ns = None
        self._nv = None
        self._use_64bit = use_64bit
        self._init = False

    def fit_u(self, traces: npt.NDArray[np.int16], x: npt.NDArray[np.uint16]):
        r"""Updates the CPA with samples of `traces` for the classes `x`.
        This method may be called multiple times.

        Parameters
        ----------
        traces :
            Array that contains the leakage traces. The array must be of
            dimension ``(n, ns)``.
        x :
            Labels for each trace. Must be of shape ``(n, nv)``.
        """
        traces = scalib.utils.clean_traces(traces, self._ns)
        x = scalib.utils.clean_labels(x, self._nv)
        if not self._init:
            self._init = True
            self._ns = traces.shape[1]
            self._nv = x.shape[1]
            self._cpa = _scalib_ext.CPA(self._nc, self._ns, self._nv, self._use_64bit)
        if x.shape[0] != traces.shape[0]:
            raise ValueError(
                f"Number of traces {traces.shape[0]} does not match size of classes array {x.shape[0]}."
            )
        # _scalib_ext uses inverted axes for x.
        # we can copy when needed, as x should be small, so this should be cheap
        # TODO: this can be non-negligible when ns is small, we can probably
        # optimize a bit.
        x = np.ascontiguousarray(x.transpose())
        with scalib.utils.interruptible():
            self._cpa.update(traces, x, get_config())

    def get_correlation(
        self,
        models: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        r"""
        Compute the correlation metric for a given model,
        which gives the leakage value for each of the ``ns`` samples, for each
        value of the intermediate variable.
        The correlation is computed for possible key values.

        Parameters
        ----------
        models :
            Array that contains the leakage models. The array must be of shape
            ``(nv, nc, ns)`` and is formatted such that the element at location
            ``[i,j,k]`` is the leakage model for ``k``-th leakage sample
            associated to the ``j``-th class of the intermediate state for the
            ``i``-th variable.

        Returns
        -------
        Correlations as an array of shape ``(nv, nc, ns)``, such that the
        element at location ``[i,j,k]`` is the correlation computed for the
        ``k``-th leakage sample of the ``i``-th variable, under the assumption
        that the key guess ``j`` is used.
        """
        if not self._init:
            raise ValueError("Need to call .fit_u at least once.")
        with scalib.utils.interruptible():
            return self._cpa.compute_cpa(models, self._kind, get_config())



from scalib.attacks import Cpa

from utils import sbox, gen_traces
import numpy as np


def generate_hw_models(nv, nc, ns):
    hw_uv = np.zeros([nv, nc, ns])
    for c in range(nc):
        hw_uv[:, c, :] = sbox[c].bit_count()
    return hw_uv.astype(np.float64)


def cpa_HW_example(pts, traces):
    """
    Perform a Correlation Power Attack against the output of the Sbox layer in
    the first round of the AES.

    pts: shape (n, 16),
        The known plaintexts.
    traces: shape (n, ns)
        The traces, of 'ns' samples, correpsonding to the execution of the
        plaintext with the unknown fixed key.

    returns:
        The key candidate maximising the correlation with an Hamming Weight leakage model.
    """
    # First, we create the models. Models map the value of the intermediate
    # variable to the leakage (represented as a lookup table).
    # Note that this step can be done in an offline first step.
    models = generate_hw_models(16, 256, traces.shape[1])
    # Next, we create the CPA object, using the XOR intermediate function: the
    # intermediate variable will be the XOR of the guessed key and the provided
    # class (plaintext).
    cpa = Cpa(256, Cpa.Xor)
    # Fit the correlation computation (this may be called multiple times with
    # different chunks of traces/plaintexts).
    cpa.fit_u(traces, pts)
    # Compute the correlation for each variable, associated to each key guess
    # and time sample.
    corr = np.abs(cpa.get_correlation(models))
    # Return the key guess maximising the correlation for every variable.
    return np.argmax(np.max(corr, axis=2), axis=1)


def main_cpa_example():
    # Parameters
    std = 2  # Noise standard deviation
    n_traces = 1000  # Amount of traces

    # First, we simulate the traces.
    print(
        f"Generate {n_traces} simulated traces (Hamming weight + Gaussian noise)"
        f"with noise standard deviation of {std}."
    )
    traces, labels = gen_traces(n_traces, std, random_key=False, random_plaintext=True)
    # Recover only the plaintext from the labels.
    pts = np.hstack([labels["p{}".format(i)][:, np.newaxis] for i in range(16)]).astype(
        np.uint16
    )

    # Perform the CPA
    key_guessed = cpa_HW_example(pts, traces)
    print(f"Key guessed: {key_guessed}")

    # Correct key comparison
    key_correct = np.array([int(labels["k{}".format(i)][0]) for i in range(16)])
    print("Correct key: {}".format(key_correct))
    if (key_guessed == key_correct).all():
        print("Attack success")
    else:
        print("Attack failure")


if __name__ == "__main__":
    print("Example HW CPA attack")
    main_cpa_example()



__all__ = [
    "preprocessing",
    "config",
    "attacks",
    "metrics",
    "modeling",
    "postprocessing",
    "tools",
    "ScalibError",
]


from .build_config import REQUIRE_X86_64_V3
from .version import version as __version__  # noqa: F401

from ._cpu_check import support_x86_64_v3

if REQUIRE_X86_64_V3 and not all(s for _, s in support_x86_64_v3()):
    # mf = ", ".join(f for f in X86_64_V3_FEATURES if f not in cpu["flags"])
    raise ImportError(
        "This SCALib build requires x86-64-v3 level processor features, which "
        + "are not all supported by this CPU or OS."
        + "See https://github.com/simple-crypto/SCALib/blob/main/README.rst "
        + "for compiling a SCALib adapted to your CPU."
        + f"Missing features: {[f for f, a in support_x86_64_v3() if not a]}."
        #    + f"(Missing features: {mf})."
    )

from scalib._scalib_ext import ScalibError



r"""Estimation of the rank of a true key given likelihood of sub-keys.

Rank estimation estimates the rank of the full (e.g. 128-bit) key based on a
score for each value of its (e.g 8-bit) sub-keys. The scores must be additive
and positive, with lower score meaning lower rank (e.g. negative
log-likelihoods).

The `rank_accuracy` function allows to specify the desired precision of the
bound in bits. It is the high-level, and its usage is recommended over the
lower-level `rank_nbin` function.
That function gives direct, lower-level control to the core algorithm, allowing
to specify the number of bins in the histograms (whereas `rank_accuracy` tunes
this parameter automatically).

By default, the `rank_accuracy` function uses scaled histograms for convolution. 
Unlike the standard histogram approach, this method scales down the bins before
each convolution by a factor chosen to prevent exceeding a predefined limit.
After the convolution, the histogram is scaled back to restore its original magnitude.
This ensures correct functionality by avoiding negative values due to f64 overflows.
Additionally, two histograms are used — one for the lower bound and one for the upper
bound — to mitigate rounding errors and maintain precision.

Examples
--------

>>> from scalib.postprocessing import rank_accuracy
>>> import numpy as np
>>> # define the correct key
>>> key = np.random.randint(0,256,16)
>>> # Derive the score for each key byte
>>> scores = np.ones((16,256)) * 1E-5
>>> scores[np.arange(16),key] = 1
>>> # Compute the full key rank (correct key must have rank 1).
>>> (rmin,r,rmax) = rank_accuracy(-np.log(scores),key)
>>> assert r == 1


Reference
---------

.. currentmodule:: scalib.postprocessing.rankestimation

.. autosummary::
    :toctree:
    :nosignatures:
    :recursive:

    rank_accuracy
    rank_nbin

Notes
-----
The rank estimation algorithm is based on :footcite:p:`rankest_histograms`,
with the following
optimization: computation of histogram bins with higher score than the expected
key is skipped, since it has no impact on the final rank.

References
----------

.. footbibliography::
"""

__all__ = ["rank_nbin", "rank_accuracy", "HAS_NTL"]
import math

from scalib import _scalib_ext
from scalib.config import get_config
import scalib.utils

HAS_NTL: bool = _scalib_ext.HAS_NTL


def rank_nbin(costs, key, nbins, method="hist"):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Warning: this is a low-level function, you probably want to use
    `rank_accuracy` instead.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape ``(ns,nc)`` where
        ``ns`` is the number of sub-keys, ``nc`` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape ``(ns,)``.
    nbins : int
        Number of bins for each of the distributions.
    method : string
        Method used to estimate the rank. Can be the following:

        * "scaledhist": using two scaled histograms (default).
        * "hist": using histograms.
        * "naive": enumerate possible keys (very slow).
        * "histbignum": using NTL library, allows better precision.

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the estimated key rank.
            - **rmax** is an upper bound for the key rank.
    """
    with scalib.utils.interruptible():
        return _scalib_ext.rank_nbin(
            costs, key, nbins, _choose_merge_value(costs), method, get_config()
        )


def rank_accuracy(costs, key, acc_bit=1.0, method="scaledhist", max_nb_bin=2**26):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape ``(ns,nc)`` where
        ``ns`` is the number of sub-keys, ``nc`` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape ``(ns,)``.
    acc_bit : f64, default: 1.0
        Expected log2 accuracy for the key rank estimation.

        The algorithms attempts to get a result such that `rmax/rmin < 2^acc_bit`,
        but may not achieve this if the result is computationally intractable.
        In such a case, rank estimate and rank bounds are still returned, but
        the inequality may be violated.
    method : string
        Method used to estimate the rank. Can be the following:

        * "scaledhist": using two scaled histograms (default).
        * "hist": using histograms.
        * "ntl": using NTL library, allows better precision.
    max_nb_bin : int, default: :math:`2^{26}`
        Maximum number of bins to use.
        This fixes an upper bound on the computational cost of the algorithm
        (if too low, the requested accuracy might not be reached).

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the estimated key rank.
            - **rmax** is an upper bound for the key rank.
    """
    with scalib.utils.interruptible():
        return _scalib_ext.rank_accuracy(
            costs,
            key,
            2.0**acc_bit,
            _choose_merge_value(costs),
            method,
            max_nb_bin,
            get_config(),
        )


def _choose_merge_value(costs):
    """The merge parameter is the number of sub-keys to merge in a
    brute-force manner before computing histograms. Merging may improve
    accuracy at the expense of running time.
    Here we limit sub-histograms to :math:`2^{16}` values.
    """
    return min(len(costs), max(1, int(16 / math.log2(max(len(c) for c in costs)))))



import os

from setuptools import setup
from setuptools_rust import Binding, RustExtension

# Ensure these are present (in case we are not using PEP-518 compatible build
# system).
import setuptools_scm


def env_true(env):
    return env == "1"


noflags = env_true(os.environ.get("SCALIB_NOFLAGS"))
portable = not noflags and env_true(os.environ.get("SCALIB_PORTABLE"))
x86_64_v3 = not noflags and env_true(os.environ.get("SCALIB_X86_64_V3"))

if portable and x86_64_v3:
    raise ValueError("Cannot have both SCALIB_PORTABLE and SCALIB_X86_64_V3.")

# We check only for X86_64_V3, as this is the CI default, otherwise we assume local
# builds.
with open("src/scalib/build_config.py", "w") as f:
    f.write(f"REQUIRE_X86_64_V3 = {x86_64_v3}\n")

rustflags = os.environ.get("RUSTFLAGS", "")

if noflags or portable:
    pass
elif x86_64_v3:
    rustflags += " -C target-cpu=x86-64-v3"
else:
    rustflags += " -C target-cpu=native"

print(f"Build config: {noflags=} {portable=} {x86_64_v3=} {rustflags=}.")

scalib_features = ["pyo3/abi3"]

if env_true(os.environ.get("SCALIB_BLIS")):
    scalib_features.append("blis")
if env_true(os.environ.get("SCALIB_NTL")):
    scalib_features.append("ntl")

extension_args = dict(
    binding=Binding.PyO3,
    py_limited_api=True,
    rust_version=">=1.83",
)

setup(
    rust_extensions=[
        RustExtension(
            "scalib._scalib_ext",
            path="src/scalib_ext/scalib-py/Cargo.toml",
            features=scalib_features,
            env=os.environ | {"RUSTFLAGS": rustflags},
            **extension_args,
        ),
        RustExtension(
            "scalib._cpu_check",
            path="src/scalib_ext/cpu-check/Cargo.toml",
            **extension_args,
        ),
    ],
    options={"bdist_wheel": {"py_limited_api": "cp310"}},
)



import hashlib
import inspect
import numpy as np


def get_rng(**args):
    """
    Hash caller name (i.e. test name) to get the rng seed.
    args are also hashed in the seed.
    """
    # Use a deterministic hash (no need for cryptographic robustness, but
    # python's hash() is randomly salted).
    seed = hashlib.sha256((repr(args) + inspect.stack()[1][3]).encode()).digest()
    return np.random.Generator(np.random.PCG64(list(seed)))
